import pytest
from prejudice.errors import ConstraintError, ConstraintsErrors


def test_error():
    error = ConstraintError('test')
    assert error.message == 'test'
    assert error == ConstraintError('test')
    assert not error == 'test'

    with pytest.raises(AttributeError) as exc:
        error.message = 'I am immutable'


def test_errors():
    error1 = ConstraintError('test 1')
    error2 = ConstraintError('test 2')

    errors = ConstraintsErrors(error1, error2)
    assert len(errors) == 2
    assert errors == ConstraintsErrors(error1, error2)
    assert errors == [error1, error2]
    assert not errors == (error1, error2)

    assert errors.json() == (
        '''[{"message": "test 1"}, {"message": "test 2"}]'''
    )



from functools import wraps
from typing import Optional
from prejudice.errors import ConstraintError, ConstraintsErrors
from prejudice.types import Predicates, PredicateErrorHandler


def with_predicates(
        predicates: Predicates,
        handler: Optional[PredicateErrorHandler] = None):
    def predication_wrapper(func):
        @wraps(func)
        def assert_predicates(*args, **kwargs):
            for predicate in predicates:
                try:
                    predicate(*args, **kwargs)
                except (ConstraintError, ConstraintsErrors) as exc:
                    if handler is not None:
                        return handler(exc)
                    raise
            return func(*args, **kwargs)
        return assert_predicates
    return predication_wrapper



import json
from http import HTTPStatus
from typing import List, Iterable, TypeVar


HTTPCode = TypeVar('HTTPCode', HTTPStatus, int)


cdef class ConstraintError(Exception):

    cdef readonly str message
    cdef object __weakref__ # enable weak referencing support

    def __init__(self, message: str):
        self.message = message

    def __eq__(self, error):
        if isinstance(error, self.__class__):
            return error.message == self.message
        return False

    def to_dict(self):
        return {
            "message": self.message
        }


cdef class HTTPConstraintError(ConstraintError):

    cdef readonly unsigned int status

    def __init__(self, message: str, status: HTTPCode = 400):
        self.message = message
        self.status = HTTPStatus(status)

    def __eq__(self, error):
        if isinstance(error, self.__class__):
            return (
                error.message == self.message and
                self.status == error.status
            )
        return False


cdef class ConstraintsErrors(Exception):

    cdef public list errors

    def __init__(self, *errors: ConstraintError):
        self.errors: List[ConstraintError] = list(errors)

    def __iter__(self):
        return iter(self.errors)

    def __len__(self):
        return len(self.errors)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.errors == other.errors
        elif isinstance(other, Iterable):
            return self.errors == other
        return False

    cpdef str json(self):
        return json.dumps([e.to_dict() for e in self.errors])



import typing as t
from prejudice import errors


Error = t.Union[
    errors.ConstraintError,
    errors.ConstraintsErrors
]
Predicate = t.Callable[..., t.Any]
Predicates = t.Iterable[Predicate]
PredicateErrorHandler = t.Callable[[Error], t.Any]



from prejudice.errors import ConstraintError, ConstraintsErrors
from prejudice.types import Predicates


cpdef _resolve(tuple validators, tuple args, dict namespace):
    cdef list errors = []
    for validator in validators:
        try:
            validator(*args, **namespace)
        except ConstraintError as exc:
            errors.append(exc)
        except ConstraintsErrors as exc:
            errors.extend(exc.errors)
    if errors:
        return ConstraintsErrors(*errors)


def resolve_constraints(validators: Predicates, *args, **kwargs):
    return _resolve(tuple(validators), args, kwargs)



from dataclasses import dataclass
from prejudice.errors import ConstraintError, ConstraintsErrors
from prejudice.validators import Or
from prejudice.utils import resolve_constraints
from prejudice.types import Predicate


@dataclass
class Document:
    id: str
    body: str = ''
    content_type: str = 'text/plain'


def non_empty_document(item):
    """Implementation of a validator/predicate
    """
    if not item.body:
        raise ConstraintError('Body is empty.')


def must_be_real_document(item):
    """Implementation of a validator/predicate
    """
    if item.id == 'test':
        raise ConstraintError('The item must not be a test document.')


class ContentType:

    def __init__(self, content_type):
        self.ct = content_type

    def __call__(self, item):
        if item.content_type != self.ct:
            raise ConstraintError(
                f'Expected {self.ct}, got {item.content_type}.')


class TestFunctionValidators:

    def test_single_resolution(self):
        document = Document(id='test', body='This is a test.')
        errors = resolve_constraints([non_empty_document], document)
        assert errors is None

        document = Document(id='test', body='')
        errors = resolve_constraints([non_empty_document], document)
        assert isinstance(errors, ConstraintsErrors)
        assert list(errors) == [ConstraintError('Body is empty.')]


    def test_multiple_resolution(self):
        document = Document(id='test', body='This is a test.')
        errors = resolve_constraints(
            [non_empty_document, must_be_real_document], document)
        assert list(errors) == [
            ConstraintError('The item must not be a test document.')
        ]

        document = Document(id='test')
        errors = resolve_constraints(
            [non_empty_document, must_be_real_document], document)
        assert list(errors) == [
            ConstraintError('Body is empty.'),
            ConstraintError('The item must not be a test document.')
        ]


class TestValidators:

    def test_single_resolution(self):
        document = Document(id='test', body='This is a test.')
        assert document.content_type == 'text/plain'
        errors = resolve_constraints([ContentType('text/plain')], document)
        assert errors is None

        errors = resolve_constraints([ContentType('text/html')], document)
        assert list(errors) == [
            ConstraintError('Expected text/html, got text/plain.')
        ]

    def test_multiple_resolution(self):
        document = Document(id='test', body='This is a test.')
        assert document.content_type == 'text/plain'
        errors = resolve_constraints([
            ContentType('text/plain'),
            must_be_real_document
        ], document)
        assert list(errors) == [
            ConstraintError('The item must not be a test document.')
        ]

        errors = resolve_constraints([
            ContentType('text/html'),
            must_be_real_document
        ], document)
        assert list(errors) == [
            ConstraintError('Expected text/html, got text/plain.'),
            ConstraintError('The item must not be a test document.')
        ]

        document = Document(
            id='not_test',
            body='<html></html>',
            content_type='text/html'
        )
        errors = resolve_constraints([
            ContentType('text/html'),
            must_be_real_document
        ], document)
        assert errors is None


class TestOr:

    def test_typing(self):
        import typing as t

        _or = Or()
        assert isinstance(_or, tuple)
        assert isinstance(_or, t.Callable)
        assert _or == tuple()

    def test_empty_or(self):
        _or = Or()
        assert _or() is None

    def test_basic_usage(self):
        import pytest

        _or = Or((ContentType('text/plain'), ContentType('text/html')))
        assert isinstance(_or, tuple)

        document = Document(id='test')
        assert _or(document) is None

        document = Document(id='test', content_type='application/json')
        with pytest.raises(ConstraintsErrors) as exc:
            _or(document)
            assert list(exc.value) == [
                ConstraintError('Expected text/plain, got application/json.'),
                ConstraintError('Expected text/html, got application/json.'),
            ]

    def test_stacked_or(self):
        import pytest

        _or = Or((
            ContentType('text/plain'),
            Or((ContentType('text/html'), non_empty_document))
        ))
        document = Document(id='test')
        assert _or(document) is None

        document = Document(id='test', content_type='application/json')
        with pytest.raises(ConstraintsErrors) as exc:
            _or(document)

        assert list(exc.value) == [
            ConstraintError('Expected text/plain, got application/json.'),
            ConstraintError('Expected text/html, got application/json.'),
            ConstraintError('Body is empty.'),
        ]

        _or = Or((
            ContentType('text/plain'),
            Or((
                Or((ContentType('text/html'), must_be_real_document)),
                non_empty_document))
            ))
        document = Document(id='123', content_type='application/json')
        assert _or(document) is None

        document = Document(id='test', content_type='application/json')
        with pytest.raises(ConstraintsErrors) as exc:
            _or(document)

        assert list(exc.value) == [
            ConstraintError('Expected text/plain, got application/json.'),
            ConstraintError('Expected text/html, got application/json.'),
            ConstraintError('The item must not be a test document.'),
            ConstraintError('Body is empty.'),
        ]

    def test_resolve_Or(self):
        _or = Or((
            ContentType('text/plain'),
            Or((ContentType('text/html'), non_empty_document))
        ))

        document = Document(id='test', content_type='application/json')
        errors = resolve_constraints(_or, document)
        assert errors == ConstraintsErrors(
            ConstraintError('Expected text/plain, got application/json.'),
            ConstraintError('Expected text/html, got application/json.'),
            ConstraintError('Body is empty.')
        )



import pytest
from prejudice.errors import ConstraintError, ConstraintsErrors
from prejudice.decorator import with_predicates


class Document:

    def __init__(self, id):
        self.id = id


def assert_not_empty(item):
    if item is None:
        raise ConstraintError('No null object allowed.')


def validate_document(doc):
    errors = []
    if not doc:
        errors.append(ConstraintError('Document is empty.'))
    if not getattr(doc, 'id', None):
        errors.append(ConstraintError('Document does not have an id.'))
    if errors:
        raise ConstraintsErrors(*errors)


def test_simple_decorator():

    @with_predicates([validate_document])
    def handle_doc(doc):
        return True


    with pytest.raises(ConstraintsErrors) as err:
        handle_doc(None)

    assert err.value.json() == (
        '[{"message": "Document is empty."}, '
        '{"message": "Document does not have an id."}]'
    )

    assert handle_doc(Document('abc')) is True


    @with_predicates([assert_not_empty, validate_document])
    def handle_doc(doc):
        return True

    with pytest.raises(ConstraintError) as err:
        handle_doc(None)

    assert err.value.message == "No null object allowed."



import os
from setuptools import setup, Extension, find_packages

version = "0.1"

install_requires = [
]

test_requires = [
    'pytest',
]


setup(
    name='prejudice',
    version=version,
    author='Souheil CHELFOUH',
    author_email='trollfot@gmail.com',
    url='https://github.com/HorsemanWSGI/prejudice',
    download_url='http://pypi.python.org/pypi/prejudice',
    description='Python/Cython predicate/guard/validation system.',
    long_description=(open("README.rst").read() + "\n" +
                      open(os.path.join("docs", "HISTORY.rst")).read()),
    license='ZPL',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    ext_modules=[
        Extension(
            "prejudice.errors",
            ["src/prejudice/errors.c"],
            extra_compile_args=["-O3"],  # Max optimization when compiling.
        ),
        Extension(
            "prejudice.utils",
            ["src/prejudice/utils.c"],
            extra_compile_args=["-O3"],  # Max optimization when compiling.
        ),
    ],
    install_requires=install_requires,
    extras_require={
        'test': test_requires,
    },
)



import pytest
from prejudice.errors import (
    ConstraintError, HTTPConstraintError, ConstraintsErrors,
)


def test_httperror_malformed():
    with pytest.raises(TypeError) as exc:
        HTTPConstraintError()

    assert str(exc.value) == (
        "__init__() takes at least 1 positional argument (0 given)"
    )

    with pytest.raises(TypeError) as exc:
        HTTPConstraintError(status=303)

    assert str(exc.value) == (
        "__init__() takes at least 1 positional argument (0 given)"
    )

    with pytest.raises(ValueError) as exc:
        HTTPConstraintError(message='test', status='abc')

    assert str(exc.value) == "'abc' is not a valid HTTPStatus"


def test_httperror():
    error = HTTPConstraintError(message='test')
    assert error.status == 400

    error = HTTPConstraintError(message='test', status=404)
    assert error == HTTPConstraintError(message='test', status=404)

    with pytest.raises(AttributeError) as exc:
        error.message = 'I am immutable'

    with pytest.raises(AttributeError) as exc:
        error.status = 200

    error = HTTPConstraintError('test', 400)
    assert error == HTTPConstraintError(message='test', status=400)


def test_httperrors():
    error1 = ConstraintError('test 1')
    error2 = HTTPConstraintError('test 2', 400)

    errors = ConstraintsErrors(error1, error2)
    assert len(errors) == 2
    assert errors == ConstraintsErrors(error1, error2)
    assert errors == [error1, error2]
    assert not errors == (error1, error2)

    assert errors.json() == (
        '''[{"message": "test 1"}, {"message": "test 2"}]'''
    )



from typing import Tuple
from prejudice.errors import ConstraintError, ConstraintsErrors
from prejudice.types import Predicate


class Or(Tuple[Predicate]):

    def __call__(self, *args, **namespace):
        errors = []
        for validator in self:
            try:
                validator(*args, **namespace)
            except ConstraintError as exc:
                errors.append(exc)
            except ConstraintsErrors as exc:
                errors.extend(exc.errors)
            else:
                # The validator was successful, stop.
                return
        if errors:
            raise ConstraintsErrors(*errors)



from .errors import ConstraintError, HTTPConstraintError, ConstraintsErrors
from .validators import Or
from .utils import resolve_constraints
from .decorator import with_predicates


__all__ = [
    'ConstraintError',
    'ConstraintsErrors',
    'HTTPConstraintError',
    'Or',
    'resolve_constraints',
    'with_predicates'
]
